open Parser
(* type token *)
type token = Id of Strng.t | Key of Strng.t | LP | RP | Dot | Lam
(* *)
module Lexer = struct
(* Lexer.keywords *)
let keywords = Lst.map Strng.of_string [
 "if";"then";"else";"let";"in";"="
]
(* Lexer.is_key *)
let is_key id = Lst.mem id keywords
(* Lexer.idorkey *)
let idorkey =
  many1 (noneof "\\.()\n\t\r ") >>= fun cs ->
  if is_key cs then return(Key cs)
               else return(Id cs)
(* Lexer.lpar *)
let lpar = char '(' >>= fun _ -> return LP
(* Lexer.rpar *)
let rpar = char ')' >>= fun _ -> return RP
(* Lexer.dot *)
let dot  = char '.' >>= fun _ -> return Dot
(* Lexer.lam *)
let lam  = char '\\' >>= fun _ -> return Lam
(* Lexer.lex *)
let lex p =
  p      >>= fun r ->
  spaces >>= fun _ ->
  return r
(* Lexer.token *)
let token = lex(lpar <|> rpar <|> dot <|> lam <|> idorkey)
(* Lexer.tokenize *)
let tokenize =
  spaces     >>= fun _  ->
  many token >>= fun ts ->
  eoi        >>= fun _  ->
  return ts
(* *)
end
(* type t *)
type t =
  | Var of Strng.t
  | Con of Strng.t
  | App of (t * t)
  | Abs of (Strng.t * t)
  | Let of (Strng.t * t * t)
  | Ite of (t * t * t)
(* primitives : Strng.t list *)
let primitives = Lst.map Strng.of_string [
 "+";"-";"*";"/";"<";">";"=";"0";"1";"2";"3";
 "not";"true";"false";
 "pair";"fst";"snd";
 "[]";"cons";"hd";"tl";
 "Y";
]
(* is_prim : Strng.t -> bool *)
let is_prim id = Lst.mem id primitives
(* *)
module Parser = struct
(* Parser.key *)
let key s = token(function
  | Key k when k = Strng.of_string s -> Some k
  | _ -> None)
(* Parser.var *)
let var = token(function
  | Id x when not(is_prim x) -> Some x
  | _ -> None)
(* Parser.con *)
let con = token(function
  | Id x when is_prim x -> Some x
  | _ -> None)
(* Parser.lpar *)
let lpar = token(function LP -> Some() | _ -> None)
(* Parser.rpar *)
let rpar = token(function RP -> Some() | _ -> None)
(* Parser.dot *)
let dot  = token(function Dot -> Some() | _ -> None)
(* Parser.lam *)
let lam  = token(function Lam -> Some() | _ -> None)
(* Parser.mk_abs *)
let rec mk_abs body = function
  | []    -> body
  | x::xs -> Abs(x,mk_abs body xs)
(* Parser.e *)
let rec e() =
  g() >>= fun e ->
  f e
(* Parser.f *)
and f e1 = (
  g()   >>= fun e2 ->
  f(App(e1,e2))
) <|> return e1
(* Parser.ite *)
and ite() = (
  key "if"   >>= fun _  ->
  e()        >>= fun e1 ->
  key "then" >>= fun _  ->
  e()        >>= fun e2 ->
  key "else" >>= fun _  ->
  e()        >>= fun e3 ->
  return(Ite(e1,e2,e3))
)
(* Parser.letin *)
and letin() = (
  key "let" >>= fun _  ->
  var       >>= fun x  ->
  key "="   >>= fun _  ->
  e()       >>= fun e1 ->
  key "in"  >>= fun _  ->
  e()       >>= fun e2 ->
  return(Let(x,e1,e2))
)
(* Parser.abs *)
and abs() = (
  lam       >>= fun _  ->
  many1 var >>= fun xs ->
  dot       >>= fun _  ->
  e()       >>= fun e  ->
  return(mk_abs e xs)
)
(* Parser.g *)
and g() = (
  var >>= fun x ->
  return(Var x)
) <|> (
  con >>= fun x ->
  return(Con x)
) <|> (
  lpar >>= fun _   ->
  e()  >>= fun exp ->
  rpar >>= fun _   ->
  return exp
) <|> ite() <|> letin() <|> abs()
(* Parser.expression *)
let expression =
  e() >>= fun e ->
  eoi >>= fun _ ->
  return e
(* *)
end
(* of_string *)
let of_string s =
  let input = Strng.of_string s in
  match parse Lexer.tokenize input with
    | None    -> failwith "lexing error"
    | Some ts ->
      match parse Parser.expression ts with
        | None   -> failwith "parse error"
        | Some e -> e
(* toplevel_printer : Format.formatter -> t -> unit *)
let rec toplevel_printer fmt = function
 | Var x | Con x -> Format.fprintf fmt "%s" (Strng.to_string x)
 | Let(x,e1,e2) -> Format.fprintf fmt "let %s = %a in %a"
  (Strng.to_string x)
  toplevel_printer e1
  toplevel_printer e2
 | Ite(e1,e2,e3) -> Format.fprintf fmt "if %a then %a else %a"
  toplevel_printer e1
  toplevel_printer e2
  toplevel_printer e3
 | Abs(x,e) -> Format.fprintf fmt "\\%s%a" (Strng.to_string x)
  print_abs e
 | App((Abs _ as e1),(App _ as e2)) | App((Abs _ as e1),(Abs _ as e2)) ->
  Format.fprintf fmt "(%a) (%a)"
   toplevel_printer e1
   toplevel_printer e2
 | App((Abs _ as e1),e2) -> Format.fprintf fmt "(%a) %a"
  toplevel_printer e1
  toplevel_printer e2
 | App(e1,(App _ as e2)) | App(e1,(Abs _ as e2)) ->
  Format.fprintf fmt "%a (%a)"
   toplevel_printer e1
   toplevel_printer e2
 | App(e1,e2) -> Format.fprintf fmt "%a %a"
  toplevel_printer e1
  toplevel_printer e2
(* print_abs *)
and print_abs fmt = function
  | Abs(x,e) -> Format.fprintf fmt " %s%a" (Strng.to_string x) print_abs e
  | e        -> Format.fprintf fmt ".%a" toplevel_printer e
(* *)
