open BinTree
(* type t *)
type 'a t = 'a BinTree.t 
(* empty : 'a t *)
let empty = Empty
(* singleton : 'a t *)
let singleton x = Node(Empty,x,Empty)
(* insert : 'a -> 'a t -> 'a t *)
let rec insert x = function
 | Empty            -> singleton x
 | Node(l,y,r) as n -> if x = y then n else (
  if x < y then Node(insert x l,y,r)
           else Node(l,y,insert x r)
 )
(* union : 'a t -> 'a t -> 'a t *)
let rec union set = function
 | Empty       -> set
 | Node(l,v,r) -> union (union (insert v set) l) r
(* remove_min : 'a t -> ('a option,'a t) *)
let rec remove_min = function
 | Empty           -> (None,Empty)
 | Node(Empty,v,r) -> (Some v,r)
 | Node(l,v,r)     ->
  let (m,l) = remove_min l in (m,Node(l,v,r))
(* remove_max : 'a t -> ('a,'a t) *)
let rec remove_max = function
 | Empty           -> failwith "empty tree"
 | Node(l,v,Empty) -> (v,l)
 | Node(l,v,r)     ->
  let (m,r) = remove_max r in (m,Node(l,v,r))
(* remove : 'a -> 'a t -> 'a t *)
let rec remove x = function
 | Empty       -> Empty
 | Node(l,y,r) -> if x < y then Node(remove x l,y,r) else (
  if x > y then Node(l,y,remove x r) else (
   if l = Empty then r
                else let (m,l) = remove_max l in Node(l,m,r)
   )
 )
(* diff : 'a t -> 'a t -> 'a t *)
let rec diff set = function
 | Empty       -> set
 | Node(l,v,r) -> diff (remove v set) (union l r)
(* of_list : 'a list -> 'a t *)
let rec of_list = function []    -> Empty
                         | x::xs -> insert x (of_list xs)
(* to_list : 'a t -> 'a list *)
let rec to_list xt = match remove_min xt with
 | (None,_)    -> []
 | (Some m,xt) -> m :: to_list xt
(* mem : 'a -> 'a t -> bool *)
let rec mem x = function
 | Empty       -> false
 | Node(l,v,r) ->
  if x = v then true else if x < v then mem x l
                                   else mem x r
(* is_empty : 'a t -> bool *)
let is_empty xt = xt = Empty
(* min : 'a t -> 'a *)
let rec min = function
 | Empty           -> failwith "empty set"
 | Node(Empty,v,_) -> v
 | Node(l,_,_)     -> min l
(* choose : 'a t -> 'a *)
let choose xt = min xt
(* unwrap : 'a option -> 'a *)
let unwrap = function None   -> failwith "empty option"
                    | Some x -> x
(* fold : ('a -> 'b -> 'b) -> 'b -> 'a t -> 'b *)
let rec fold f b ts = if ts = Empty then b else (
 let (m,ts) = remove_min ts in
 f (unwrap m) (fold f b ts)
)
(* filter : ('a -> bool) -> 'a t -> 'a t *)
let filter p ts = fold (fun x d -> if p x then insert x d else d) empty ts
(* is_singleton : 'a t -> bool *)
let is_singleton = function Node(Empty,_,Empty) -> true
                          | _                   -> false
(* fprintf : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit *)
let fprintf print_element fmt set =
 let rec fprintf_list fmt = function
  | []    -> ()
  | [x]   -> print_element fmt x
  | x::xs -> Format.fprintf fmt "%a,%a" print_element x fprintf_list xs
 in
 Format.fprintf fmt "{%a}%!" fprintf_list (to_list set)
(* toplevel_printer_int : Format.formatter -> int t -> unit *)
let toplevel_printer_int = fprintf (fun fmt i -> Format.fprintf fmt "%i" i)
(* *)
